"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides our database session interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import copy
import functools
from typing import Any, Optional

import sqlalchemy
import sqlalchemy.orm
import sqlalchemy.pool

import PyFunceble.sessions
from PyFunceble.database.credential.base import CredentialBase


class DBSession:
    """
    Provides our very own database session interface and handler.
    """

    credential: Optional[CredentialBase] = None
    current_session: sqlalchemy.orm.Session = None

    def __enter__(self) -> sqlalchemy.orm.Session:
        if PyFunceble.sessions.DB_FACTORY is not None:
            self.current_session = PyFunceble.sessions.DB_FACTORY()
        else:
            self.current_session = self.get_new_session()()

        return self.current_session

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if exc_type or exc_val or exc_tb:
            try:
                self.current_session.rollback()
            except AttributeError:
                pass

        self.close()

    def ensure_credential_is_given(func):  # pylint: disable=no-self-argument
        """
        Ensures that a credential object is set before launching the
        decorated method.

        :raise TypeError:
            When :code:`credential` is not correct.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if not isinstance(self.credential, CredentialBase):
                raise TypeError(
                    f"<self.credential> should be {CredentialBase}, "
                    f"{type(self.credential)} given."
                )

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    def execute_if_authorized(default: Any = None):  # pylint: disable=no-self-argument
        """
        Executes the decorated method only if we are authorized to process.
        Otherwise, apply the given :code:`default`.
        """

        def inner_metdhod(func):
            @functools.wraps(func)
            def wrapper(self, *args, **kwargs):
                if self.authorized:
                    return func(self, *args, **kwargs)  # pylint: disable=not-callable
                return self if default is None else default

            return wrapper

        return inner_metdhod

    @property
    def authorized(self) -> bool:
        """
        Provides the authorization to operate.
        """

        return self.credential is not None

    @execute_if_authorized(None)
    @ensure_credential_is_given
    def init_db_sessions(self) -> "DBSession":
        """
        Initiate the global session to work with.
        """

        if PyFunceble.sessions.DB_ENGINE is None:
            PyFunceble.sessions.DB_ENGINE = sqlalchemy.create_engine(
                self.credential.get_uri(), poolclass=sqlalchemy.pool.NullPool
            )

            PyFunceble.sessions.DB_FACTORY = sqlalchemy.orm.sessionmaker(
                bind=PyFunceble.sessions.DB_ENGINE,
                autoflush=True,
                autocommit=False,
                expire_on_commit=False,
            )

        return self

    @execute_if_authorized(None)
    def get_db_session(self) -> "DBSession":
        """
        Provides a new session.
        """

        session_object = DBSession()
        session_object.credential = copy.deepcopy(self.credential)

        return session_object

    @execute_if_authorized(None)
    @ensure_credential_is_given
    def get_new_session(self) -> sqlalchemy.orm.sessionmaker:
        """
        Creates and returns a new session.

        .. warning::
            This method generate a new session without any pool of connections.
        """

        engine = sqlalchemy.create_engine(
            self.credential.get_uri(), poolclass=sqlalchemy.pool.NullPool
        )

        return sqlalchemy.orm.sessionmaker(
            bind=engine, autoflush=True, autocommit=False, expire_on_commit=False
        )

    @execute_if_authorized(None)
    @ensure_credential_is_given
    def get_new_pool_session(self) -> sqlalchemy.orm.sessionmaker:
        """
        Create and return a new session.
        """

        engine = sqlalchemy.create_engine(
            self.credential.get_uri(),
        )

        return sqlalchemy.orm.sessionmaker(
            bind=engine, autoflush=True, autocommit=False, expire_on_commit=False
        )

    def close(self) -> "DBSession":
        """
        Closes the session if exists.
        """

        if self.current_session is not None:
            self.current_session.close()

            del self.current_session
            self.current_session = None

        return self

    @execute_if_authorized(None)
    def query(self, *args, **kwargs) -> Any:
        """
        Makes a query.
        """

        # pylint: disable=no-member
        with self as db_session:
            return db_session.query(*args, **kwargs)



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides our logger.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import functools
import logging
import logging.handlers
import os
import traceback
from typing import Generator, Optional, Tuple, Union

import PyFunceble.cli.storage
import PyFunceble.storage
from PyFunceble.helpers.directory import DirectoryHelper
from PyFunceble.helpers.environment_variable import EnvironmentVariableHelper


class Logger:
    """
    Provides our logging logic.

    .. warning::
        There is several way to activate the logging.

        1. Through the :code:`PYFUNCEBLE_DEBUG_ON_SCREEN` environment variable.
        2. Through the :code:`DEBUG_PYFUNCEBLE_ON_SCREEN` environment variable.
        3. Through the :code:`PYFUNCEBLE_DEBUG` environment variable.
        4. Through the :code:`DEBUG_PYFUNCEBLE` environment variable.
        5. Through the :py:meth:`PyFunceble.logger.Logger.set_activated` method.
    """

    # pylint: disable=too-many-public-methods

    # pylint: disable=line-too-long
    OWN_FORMAT: str = (
        "[%(asctime)s | %(levelname)s | %(origin_path)s:"
        "%(origin_line)s@%(origin_func)s | PPID%(process)d:%(processName)s]:\n"
        "%(message)s\n"
    )
    """
    Our very own format.
    """

    ROOT_FORMAT: str = (
        "[%(asctime)s::%(levelname)s](PID%(process)s:%(processName)s): %(message)s"
    )
    """
    The format of the root loggy.
    """

    STD_MIN_LEVEL: int = logging.INFO

    _activated: bool = False
    _min_level: int = logging.INFO
    _output_directory: Optional[str] = None

    own_formatter: logging.Formatter = logging.Formatter(OWN_FORMAT)
    root_formatter: logging.Formatter = logging.Formatter(ROOT_FORMAT)

    info_logger: Optional[logging.Logger] = None
    debug_logger: Optional[logging.Logger] = None
    warning_logger: Optional[logging.Logger] = None
    error_logger: Optional[logging.Logger] = None
    fatal_logger: Optional[logging.Logger] = None
    critical_logger: Optional[logging.Logger] = None
    sqlalchemy_logger: Optional[logging.Logger] = None

    env_var_helper: EnvironmentVariableHelper = EnvironmentVariableHelper()

    def __init__(
        self,
        *,
        activated: Optional[bool] = None,
        min_level: Optional[int] = None,
        output_dir: Optional[str] = None,
    ) -> None:

        if output_dir:
            self.output_directory = output_dir
        else:
            self.output_directory = os.path.join(
                PyFunceble.cli.storage.OUTPUT_DIRECTORY,
                PyFunceble.cli.storage.OUTPUTS.logs.directories.parent,
            )

        if activated is not None:
            self.activated = activated

        if min_level is not None:
            self.min_level = min_level
        else:
            self.guess_and_set_min_level()

    @property
    def on_screen(self) -> bool:
        """
        Provides the authorization to log on screen.
        """

        return (
            self.env_var_helper.set_name("PYFUNCEBLE_DEBUG_ON_SCREEN").exists()
            or self.env_var_helper.set_name("DEBUG_PYFUNCEBLE_ON_SCREEN").exists()
        )

    @property
    def on_file(self) -> bool:
        """
        Provides the authorization to log on file.
        """

        return (
            self.env_var_helper.set_name("PYFUNCEBLE_DEBUG").exists()
            or self.env_var_helper.set_name("DEBUG_PYFUNCEBLE").exists()
        )

    @property
    def activated(self) -> bool:
        """
        Provides the current state of the :code:`_activated` attribute.
        """

        return self._activated

    @activated.setter
    def activated(self, value: bool) -> None:
        """
        Sets the value of the activated attribute.

        :param value:
            The value to set.

        :raise TypeError:
            When the given value is not a :py:class:`bool`.
        """

        if not isinstance(value, bool):
            raise TypeError(f"<value> should be {bool}, {type(value)} given.")

        self._activated = value

    def set_activated(self, value: bool) -> "Logger":
        """
        Sets the value of the activated attribute.

        :param value:
            The value to set.
        """

        self.activated = value

        return self

    @property
    def min_level(self) -> int:
        """
        Provides the current state of the :code:`_min_level` attribute.
        """

        return self._min_level

    @min_level.setter
    def min_level(self, value: Union[int, str]) -> None:
        """
        Sets the minimum level to log.

        :param value:
            The value to set.

        :raise TypeError:
            When the given value is not a :py:class:`int` or :py:class:`str`.
        :raise ValueError:
            When the given value is not supported
        """

        if not isinstance(value, (int, str)):
            raise TypeError(f"<value> should be {int} or {str}, {type(value)} given.")

        # pylint: disable=protected-access
        if isinstance(value, int) and value not in logging._levelToName:
            raise ValueError(f"<value> is not in {list(logging._nameToLevel.keys())}")

        if isinstance(value, str):
            if value.upper() not in logging._nameToLevel:
                raise ValueError(
                    f"<value> is not in {list(logging._nameToLevel.keys())}"
                )
            value = logging._nameToLevel[value.upper()]

        self._min_level = value

        return self

    def set_min_level(self, value: Union[int, str]) -> "Logger":
        """
        Sets the minimum level to log.

        :param value:
            The value to set.
        """

        self.min_level = value

        return self

    @property
    def output_directory(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_output_directory` attribute.
        """

        if self.authorized:
            DirectoryHelper(self._output_directory).create()

        return self._output_directory

    @output_directory.setter
    def output_directory(self, value: str) -> None:
        """
        Sets the directory to write.

        Side Effect:
            Creation of the given value if it does not exists.

        :param value:
            The value to set.

        :raise TypeError:
            When the given value is not a  :py:class:`str`.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {int}, {type(value)} given.")

        self._output_directory = value

    def set_output_directory(self, value: str) -> "Logger":
        """
        Sets the directory to write.

        :param value:
            The value to set.
        """

        self.output_directory = value

        return self

    @property
    def authorized(self) -> bool:
        """
        Provides the authorization to actually log.
        """

        return (
            self.activated
            or self.on_screen
            or self.on_file
            or (
                bool(PyFunceble.storage.CONFIGURATION)
                and bool(PyFunceble.storage.CONFIGURATION.debug.active)
            )
        )

    @staticmethod
    def get_origin() -> dict:
        """
        Provides the informatioon about where the logger was triggered.

        :return:
            A tuple, which is composed of the following.

            (trigger file path, trigger line, trigger function/method name)
        """

        stackback = [
            y for x in [x.split("\n") for x in traceback.format_stack()] for y in x if y
        ]
        interest = stackback[-6].split(",")

        complete_file = interest[0].strip()[6:-1].split(os.sep)

        try:
            if complete_file[-2] != PyFunceble.storage.PROJECT_NAME:
                file = "/".join(complete_file)
            else:
                file = "/".join(complete_file[-2:])
        except IndexError:
            file = "/".join(complete_file)

        line = interest[1].strip().split()[-1].strip()
        func_name = interest[2].strip()[3:]

        if PyFunceble.storage.PROJECT_NAME in file:
            file = os.path.relpath(file)

        return {"origin_path": file, "origin_line": line, "origin_func": func_name}

    def single_logger_factory(level_name: str):  # pylint: disable=no-self-argument
        """
        Provides the general factory.

        :param level_name:
            The level to log.
        """

        def single_logger(func):
            @functools.wraps(func)
            def wrapper(self, *args, **kwargs):
                # pylint: disable=no-member, protected-access

                if (
                    self.authorized
                    and logging._nameToLevel[level_name.upper()] >= self.min_level
                ):
                    try:
                        logger = getattr(
                            getattr(self, f"{level_name.lower()}_logger"),
                            level_name.lower(),
                        )

                        if not logger:
                            self.init_loggers()
                    except AttributeError:
                        self.init_loggers()

                        logger = getattr(
                            getattr(self, f"{level_name.lower()}_logger"),
                            level_name.lower(),
                        )

                    return logger(*args, **kwargs, extra=self.get_origin())

                return func

            return wrapper

        return single_logger

    def guess_and_set_min_level(self) -> "Logger":
        """
        Tries to guess the min level from the configuration.
        """

        env_vars = ["PYFUNCEBLE_DEBUG_LVL", "PYFUNCEBLE_LOGGING_LVL"]

        env_var_helper = EnvironmentVariableHelper()
        env_var_found = False

        for env_var in env_vars:
            if env_var_helper.set_name(env_var).exists():
                self.min_level = env_var_helper.get_value()
                env_var_found = True

                break

        if not env_var_found:
            if PyFunceble.storage.CONFIGURATION:
                # pylint: disable=protected-access
                self.min_level = logging._nameToLevel[
                    PyFunceble.storage.CONFIGURATION.debug.level.upper()
                ]
            else:
                self.min_level = self.STD_MIN_LEVEL

    def guess_all_settings(self) -> "Logger":
        """
        Try to guess all settings.
        """

        to_ignore = ["guess_all_settings"]

        for method in dir(self):
            if method in to_ignore or not method.startswith("guess_"):
                continue

            getattr(self, method)()

        return self

    def destroy_loggers(self) -> "Logger":
        """
        Destroyes all existing loggers.
        """

        for logger, level in self.get_next_logger():
            logger.handlers.clear()

            delattr(self, f"{level}_logger")
            setattr(self, f"{level}_logger", None)

    def get_next_logger(self) -> Generator[None, Tuple[logging.Logger, str], None]:
        """
        Provides the next logger.
        """

        for logger_var_name in Logger.__dict__:
            if (
                not logger_var_name.endswith("_logger")
                or logger_var_name == "get_next_logger"
                or "sqlalchemy" in logger_var_name
            ):
                continue

            level = logger_var_name.split("_", 1)[0]

            if not getattr(self, logger_var_name):
                setattr(self, logger_var_name, logging.getLogger(f"PyFunceble.{level}"))

            yield getattr(self, logger_var_name), level

    def init_loggers(self) -> "Logger":
        """
        Init all our logger.
        """

        if self.authorized:
            self.destroy_loggers()

            if not self.sqlalchemy_logger:
                self.sqlalchemy_logger = logging.getLogger("sqlalchemy")

            self.sqlalchemy_logger.setLevel(self.min_level)
            self.sqlalchemy_logger.propagate = False

            for logger, level in self.get_next_logger():
                logger.propagate = False

                # pylint: disable=protected-access
                logger.setLevel(logging._nameToLevel[level.upper()])
                logger.addHandler(self.get_handler(level))

        return self

    def get_handler(
        self, level_name: str
    ) -> Optional[Union[logging.StreamHandler, logging.handlers.RotatingFileHandler]]:
        """
        Given a level name, this method provides the right handler for it!
        """

        level_name = level_name.upper()
        specials = ["SQLALCHEMY"]

        if hasattr(logging, level_name) or level_name in specials:
            if self.on_screen:
                handler = logging.StreamHandler()
            else:
                handler = logging.handlers.RotatingFileHandler(
                    os.path.join(self.output_directory, f"{level_name.lower()}.log"),
                    maxBytes=100_000_000,
                    backupCount=5,
                )

            if level_name in specials:
                handler.setLevel(self.min_level)
                handler.setFormatter(self.root_formatter)
            else:
                handler.setLevel(getattr(logging, level_name))
                handler.setFormatter(self.own_formatter)

            return handler

        return None

    @single_logger_factory("info")
    def info(self, *args, **kwargs):
        """
        Logs to the PyFunceble.info logger.
        """

    @single_logger_factory("debug")
    def debug(self, *args, **kwargs):
        """
        Logs to the PyFunceble.debug logger.
        """

    @single_logger_factory("warning")
    def warning(self, *args, **kwargs):
        """
        Logs to the PyFunceble.warning logger.
        """

    @single_logger_factory("error")
    def error(self, *args, **kwargs):
        """
        Logs to the PyFunceble.error logger.
        """

    @single_logger_factory("fatal")
    def fatal(self, *args, **kwargs):
        """
        Logs to the PyFunceble.fatal logger.
        """

    @single_logger_factory("critical")
    def critical(self, *args, **kwargs):
        """
        Logs to the PyFunceble.critical logger.
        """

    @single_logger_factory("exception")
    def exception(self, *args, **kwargs):
        """
        Logs to the PyFunceble.exception logger.
        """



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides a way to convert our internal URL.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from typing import Any

import PyFunceble.storage
from PyFunceble.converter.base import ConverterBase
from PyFunceble.utils.version import VersionUtility


class InternalUrlConverter(ConverterBase):
    """
    Converter of the internal URLs.

    .. note::
        The internal URLs are actually the URL that has nothing to
        do with what we are going to test.

        They are only relevant for the software itself.
    """

    @ConverterBase.data_to_convert.setter
    def data_to_convert(self, value: Any) -> None:
        """
        Overrites the default behavior.

        :raise TypeError:
            When the given data to convert is not :py:class:`str`
        """
        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        # pylint: disable=no-member
        super(InternalUrlConverter, self.__class__).data_to_convert.fset(self, value)

    def get_converted(self) -> str:
        """
        Provides the converted data (after conversion)
        """

        if VersionUtility(PyFunceble.storage.PROJECT_VERSION).is_dev():
            return self.data_to_convert.replace("master", "dev")
        return self.data_to_convert.replace("dev", "master")



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides our producer worker.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from typing import Any, Optional, Tuple

import PyFunceble.facility
import PyFunceble.storage
from PyFunceble.checker.status_base import CheckerStatusBase
from PyFunceble.cli.processes.workers.producer import ProducerWorker


class ChancyProducerWorker(ProducerWorker):
    """
    Provides our chancy producer worker.

    .. warning::
        This chancy producer does not provide any guarantee. The flow that keep
        PyFunceble safe are here unleashed.

        USE AT YOUR OWN RISK. GOOD LUCK!
    """

    STD_NAME: str = "pyfunceble_chancy_producer_worker"

    def target(self, consumed: Any) -> Optional[Tuple[Any, ...]]:
        if not isinstance(consumed, tuple):
            PyFunceble.facility.Logger.info(
                "Skipping latest dataset because consumed data was not a tuple."
            )
            return None

        # Just for human brain.
        test_dataset, test_result = consumed

        if not isinstance(test_dataset, dict):
            PyFunceble.facility.Logger.info(
                "Skipping because test dataset is not a dictionnary."
            )
            return None

        if self.should_we_ignore(test_result):
            PyFunceble.facility.Logger.info(
                "Ignored test dataset. Reason: No output wanted."
            )
            return None

        if not isinstance(test_result, CheckerStatusBase):
            PyFunceble.facility.Logger.info(
                "Skipping latest dataset because consumed status is not "
                "a status object.."
            )
            return None

        self.run_whois_backup(test_result)
        self.run_inactive_backup(test_dataset, test_result)
        self.run_continue_backup(test_dataset, test_result)

        if not self.should_we_block_status_file_printer(test_dataset, test_result):
            self.run_counter(test_dataset, test_result)

        test_dataset["from_chancy_producer"] = True

        return test_dataset, test_result



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides everything related to the registrar counter.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from typing import Dict, List, Optional, Union

import PyFunceble.cli.storage
from PyFunceble.cli.filesystem.json_base import FilesystemJSONBase
from PyFunceble.helpers.list import ListHelper


class RegistrarCounter(FilesystemJSONBase):
    """
    Provides our registrar stats counter.
    """

    STD_DATASET: Dict[str, int] = {
        "counter": {
            "total": 0,
        },
        "percentage": {"total": 0},
    }

    SUPPORTED_TEST_MODES: List[str] = ["AVAILABILITY"]

    SOURCE_FILE: str = PyFunceble.cli.storage.REGISTRAR_COUNTER_FILE

    @FilesystemJSONBase.fetch_dataset_beforehand
    def get_dataset_for_printer(
        self, *, limit: Optional[int] = 15
    ) -> List[Dict[str, Union[str, int]]]:
        """
        Provides the dataset that the printer may understand.

        :param limit:
            Maximum number of registrars to display.

            .. warning::
                If set to :code:`None`, all registrars will be displayed.

        :raise ValueError:
            When the current testing mode is not supported (yet?).
        """

        result = {}
        testing_mode = PyFunceble.cli.utils.testing.get_testing_mode()

        if testing_mode not in self.SUPPORTED_TEST_MODES:
            raise ValueError("<testing_mode> ({testing_mode!r}) is not supported.")

        for registrar, value in self.dataset["counter"].items():
            if registrar == "total":
                continue

            result[registrar] = {"registrar": registrar, "amount": value}

        for registrar, value in self.dataset["percentage"].items():
            if registrar == "total":
                continue

            result[registrar]["percentage"] = f"{round(value)}%"

        # Apply the right order.
        result = (
            ListHelper([y for _, y in result.items()])
            .custom_sort(key_method=lambda x: x["amount"], reverse=True)
            .subject
        )

        return result[:limit] if limit else result

    @FilesystemJSONBase.update_source_file_path_beforehand
    @FilesystemJSONBase.fetch_dataset_beforehand
    @FilesystemJSONBase.save_dataset_afterwards
    def count(self, registrar: str) -> "RegistrarCounter":
        """
        Starts the counting process.

        :param registrar:
            The registrar to count into our dataset.
        """

        if not isinstance(registrar, str):
            raise TypeError(f"<registrar> should be {str}, {type(registrar)} given.")

        if registrar not in self.dataset["counter"]:
            self.dataset["counter"][registrar] = 1
        else:
            self.dataset["counter"][registrar] += 1

        self.dataset["counter"]["total"] += 1

        self.dataset["percentage"][registrar] = (
            self.dataset["counter"][registrar] * 100
        ) / self.dataset["counter"]["total"]

        for key in self.dataset["percentage"]:
            if key in ("total", registrar):
                continue

            self.dataset["percentage"][key] = (
                self.dataset["counter"][key] * 100
            ) / self.dataset["counter"]["total"]

        self.dataset["percentage"]["total"] = sum(
            y for x, y in self.dataset["percentage"].items() if x != "total"
        )

        return self



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides everything related to the cleanup of the filesystem.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import os
from typing import List

import PyFunceble.cli.facility
import PyFunceble.cli.factory
import PyFunceble.cli.utils.testing
import PyFunceble.sessions
from PyFunceble.cli.filesystem.dir_base import FilesystemDirBase
from PyFunceble.database.sqlalchemy.all_schemas import Continue, Inactive
from PyFunceble.helpers.file import FileHelper


class FilesystemCleanup(FilesystemDirBase):
    """
    Provides the interface for the cleanup of the filesystem.
    """

    file_helper: FileHelper = FileHelper()

    def clean_database(self) -> "FilesystemCleanup":
        """
        Cleanups the uneeded data that were stored in the database.

        .. warning::
            This method cleans everything except the WHOIS records.
        """

        if (
            PyFunceble.cli.facility.CredentialLoader.is_already_loaded()
            and self.db_session
        ):
            to_clean = [Continue, Inactive]

            for orm_obj in to_clean:
                # pylint: disable=line-too-long
                with PyFunceble.cli.factory.DBSession.get_db_session() as db_session:
                    db_session.query(orm_obj).delete(synchronize_session=False)
                    db_session.commit()

                    PyFunceble.facility.Logger.info(
                        "Deleted all entries in %r.", orm_obj
                    )

    @property
    def output_files_to_delete(self) -> List[str]:
        """
        Provides the list of output files to delete.
        """

        result = []
        files_to_ignore = [".gitignore", ".keep", ".gitkeep"]

        for root, _, files in os.walk(self.get_output_basedir()):
            for file in files:
                if file in files_to_ignore:
                    continue

                result.append(os.path.join(root, file))

        return result

    def clean_output_files(self) -> "FilesystemCleanup":
        """
        Cleanups the unneeded files from the output directory.
        """

        for file in self.output_files_to_delete:
            self.file_helper.set_path(file).delete()

            PyFunceble.facility.Logger.debug("Deleted: %r.", file)

        return self

    def start(self) -> "FilesystemCleanup":
        """
        Starts the cleanup of everything unneeded.
        """

        self.clean_output_files()

        return self



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides some of our scripts.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""


import copy
import functools
import os
import pathlib
from typing import List, Optional

import PyFunceble.cli.storage
import PyFunceble.facility
import PyFunceble.storage
from PyFunceble.cli.filesystem.dir_structure.backup import DirectoryStructureBackup
from PyFunceble.helpers.command import CommandHelper
from PyFunceble.helpers.dict import DictHelper
from PyFunceble.helpers.directory import DirectoryHelper
from PyFunceble.helpers.file import FileHelper
from PyFunceble.helpers.regex import RegexHelper
from PyFunceble.utils.version import VersionUtility


class ProductionPrep:
    """
    Provides an interface for the production file modification.
    The idea is that we always have 2 branches: the `dev` and the `master`
    branch.

    We want to fix all the URL to point to the right one, so this interface
    just provides everything needed for that job.

    Another important part is the cleanup of the production environment.
    What is meant is the cleanup of the `output/` directory and the
    construction of the dir_structure file.

    .. warning::
        This class assumes that you know what you are doing. Meaning that
        you should run this only if your are developing PyFunceble.
    """

    VERSION_FILE_PATH: str = os.path.join(
        PyFunceble.storage.CONFIG_DIRECTORY,
        PyFunceble.cli.storage.DISTRIBUTED_VERSION_FILENAME,
    )
    AVAILABLE_BRANCHES: List[str] = ["dev", "master"]

    regex_helper: RegexHelper = RegexHelper()
    file_helper: FileHelper = FileHelper()
    dict_helper: DictHelper = DictHelper()
    version_utility: VersionUtility = VersionUtility(PyFunceble.storage.PROJECT_VERSION)

    version_file_content: Optional[dict] = None
    """
    A copy of the local version file.
    """

    _branch: Optional[str] = None

    previous_version: Optional[str] = None
    """
    Provides the previous version (from :code:`version_file_content`)
    """

    def __init__(self, branch: Optional[str] = None) -> None:
        self.version_file_content = self.dict_helper.from_yaml_file(
            self.VERSION_FILE_PATH
        )

        self.previous_version = copy.deepcopy(
            self.version_file_content["current_version"]
        )

        if branch is not None:
            self.branch = branch

    def ensure_branch_is_given(func):  # pylint: disable=no-self-argument
        """
        Ensures that the branch is given before running the decorated method.

        :raise TypeError:
            When the :code:`self.branch` is not set.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if not isinstance(self.branch, str):
                raise TypeError(
                    f"<self.branch> should be {str}, " f"{type(self.branch)} given."
                )

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    @property
    def branch(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_branch` attribute.
        """

        return self._branch

    @branch.setter
    def branch(self, value: str) -> None:
        """
        Sets the branch to act with.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        self._branch = value

    def set_branch(self, value: str) -> "ProductionPrep":
        """
        Sets the branch to act with.

        :param value:
            The value to set.
        """

        self.branch = value

        return self

    def should_be_deprecated(self, previous_version: str) -> bool:
        """
        Checks if we should deprecates the current version.
        """

        splitted = self.version_utility.get_splitted(previous_version)[0]
        local_splitted = self.version_utility.get_splitted(
            self.version_utility.local_version
        )[0]

        for index, version_part in enumerate(splitted[:2]):
            if int(version_part) < int(local_splitted[index]):
                return True

        return False

    @ensure_branch_is_given
    def update_urls(self, file: str) -> "ProductionPrep":
        """
        Updates the common URLS which are in the given file.

        :param file:
            The file to work with.

        :raise FileNotFoundError:
            When the given :code:`file` is not found.
        """

        if self.branch == "dev":
            regexes = [
                (r"PyFunceble\/%s\/" % "master", "PyFunceble/%s/" % "dev"),
                ("=%s" % "master", "=%s" % "dev"),
                (r"/{1,}en\/%s" % "latest", "/en/%s" % "dev"),
                (r"\/pyfunceble-dev.png", "/pyfunceble-%s.png" % "dev"),
                (r"\/project\/pyfunceble$", "/project/pyfunceble-%s" % "dev"),
                (
                    r"\/badge\/pyfunceble(/month|/week|)$",
                    "/badge/pyfunceble-%s\\1" % "dev",
                ),
                (r"\/blob\/%s\/" % "master", "/blob/%s/" % "dev"),
                (r"\/pypi\/v\/pyfunceble\.png$", "/pypi/v/pyfunceble-%s.png" % "dev"),
                (r"\/(logo|graphmls|gifs\/raw)\/%s\/" % "master", "/\\1/%s/" % "dev"),
                (r"\/(PyFunceble\/tree)\/%s" % "master", "/\\1/%s" % "dev"),
            ]
        elif self.branch == "master":
            regexes = [
                (r"PyFunceble\/%s\/" % "dev", "PyFunceble/%s/" % "master"),
                ("=%s" % "dev", "=%s" % "master"),
                (r"/{1,}en\/%s" % "dev", "/en/%s" % "latest"),
                (r"\/pyfunceble-dev.png", "/pyfunceble-dev.png"),
                (r"\/project\/pyfunceble-%s$" % "dev", "/project/pyfunceble"),
                (
                    r"\/badge\/pyfunceble-%s(/month|/week|)$" % "dev",
                    "/badge/pyfunceble\\1",
                ),
                (r"\/blob\/%s\/" % "dev", "/blob/%s/" % "master"),
                (
                    r"\/pypi\/v\/pyfunceble-%s\.png$" % "dev",
                    "/pypi/v/pyfunceble.png",
                ),
                (r"\/(logo|graphmls|gifs\/raw)\/%s\/" % "dev", "/\\1/%s/" % "master"),
                (r"\/(PyFunceble\/tree)\/%s" % "dev", "/\\1/%s" % "master"),
            ]
        else:
            regexes = {}

        self.file_helper.set_path(file)

        PyFunceble.facility.Logger.info(
            "Started to update our URL into %r", self.file_helper.path
        )

        if not self.file_helper.exists():
            raise FileNotFoundError(self.file_helper.path)

        to_update = self.file_helper.read()

        for regex, replacement in regexes:
            to_update = self.regex_helper.set_regex(regex).replace_match(
                to_update, replacement, multiline=True
            )

        self.file_helper.write(to_update, overwrite=True)

        PyFunceble.facility.Logger.info(
            "Finished to update our URL into %r", self.file_helper.path
        )

        return self

    def update_docs_urls(self) -> "ProductionPrep":
        """
        Updates all URL in the documentation files.
        """

        to_ignore = ["they-use-d-it.rst"]

        self.update_urls(
            os.path.join(PyFunceble.storage.CONFIG_DIRECTORY, "README.rst")
        )

        for root, _, files in os.walk(
            os.path.join(PyFunceble.storage.CONFIG_DIRECTORY, "docs")
        ):
            for file in files:
                if not file.endswith(".rst"):
                    continue

                full_path = os.path.join(root, file)

                if any(x in full_path for x in to_ignore):
                    continue

                self.update_urls(os.path.join(root, file))

    @staticmethod
    def update_code_format() -> "ProductionPrep":
        """
        Updates the format of the source code using black.
        """

        # pylint: disable=import-outside-toplevel, import-error
        import black
        import isort

        def format_file(file: str, isortconfig: isort.settings.Config) -> None:
            """
            Formats the given file using black.

            :param file:
                The file to format.
            :parm isortconfig:
                The configuration to apply while sorting the imports.
            """

            isort.api.sort_file(pathlib.Path(file), config=isortconfig)

            black.format_file_in_place(
                pathlib.Path(file),
                fast=False,
                mode=black.Mode(),
                write_back=black.WriteBack.YES,
            )

            PyFunceble.facility.Logger.info("Update format of %r", file)

        isort_config = isort.settings.Config(settings_file="setup.cfg")

        files = [
            os.path.join(PyFunceble.storage.CONFIG_DIRECTORY, "setup.py"),
        ]

        for file in files:
            format_file(file, isort_config)

        for root, _, files in os.walk(
            os.path.join(
                PyFunceble.storage.CONFIG_DIRECTORY, PyFunceble.storage.PROJECT_NAME
            )
        ):
            if "__pycache__" in root:
                continue

            for file in files:
                if not file.endswith(".py"):
                    continue

                format_file(os.path.join(root, file), isort_config)

        for root, _, files in os.walk(
            os.path.join(PyFunceble.storage.CONFIG_DIRECTORY, "tests")
        ):
            if "__pycache__" in root:
                continue

            for file in files:
                if not file.endswith(".py"):
                    continue

                format_file(os.path.join(root, file), isort_config)

    @staticmethod
    def update_documentation() -> "ProductionPrep":
        """
        Updates the code documentation.

        :raise RuntimeError:
            When one of the wanted directory is not found.
        """

        PyFunceble.facility.Logger.info(
            "Started to update and generate the documentation.",
        )

        docs_dir_helper = DirectoryHelper("docs")
        source_code_dir_helper = DirectoryHelper("PyFunceble")

        if not docs_dir_helper.exists():
            raise RuntimeError(f"{docs_dir_helper.realpath!r} not found.")

        if not source_code_dir_helper.exists():
            raise RuntimeError(f"{source_code_dir_helper.realpath!r} not found.")

        header = "Code Documentation"
        source_code_destination = os.path.join(docs_dir_helper.realpath, "code")

        CommandHelper(
            f"sphinx-apidoc -d 5 -f -H {header!r} -o "
            f"{source_code_destination!r} {source_code_dir_helper.realpath}"
        ).execute(raise_on_error=True)

        docs_destination = os.path.join(docs_dir_helper.realpath, "_build", "html")

        CommandHelper(
            f"sphinx-build -a -Q {docs_dir_helper.realpath!r} {docs_destination!r}"
        ).execute(raise_on_error=False)

        PyFunceble.facility.Logger.info(
            "Finished to update and generate the documentation.",
        )

    def update_code_urls(self) -> "ProductionPrep":
        """
        Updates all URL in the source code.
        """

        to_ignore = [
            ".gitignore",
            ".keep",
        ]

        self.update_urls(os.path.join(PyFunceble.storage.CONFIG_DIRECTORY, "setup.py"))

        for root, _, files in os.walk(
            os.path.join(
                PyFunceble.storage.CONFIG_DIRECTORY, PyFunceble.storage.PROJECT_NAME
            )
        ):
            if "__pycache__" in root:
                continue

            for file in files:
                if file in to_ignore:
                    continue

                self.update_urls(os.path.join(root, file))

        for root, _, files in os.walk(
            os.path.join(PyFunceble.storage.CONFIG_DIRECTORY, "tests")
        ):
            if "__pycache__" in root:
                continue

            for file in files:
                if file in to_ignore:
                    continue

                self.update_urls(os.path.join(root, file))

    @ensure_branch_is_given
    def update_setup_py(self) -> "ProductionPrep":
        """
        Updates content of :code:`setup.py`.

        :raise FileNotFoundError:
            When the :code:`setup.py` file does not exists.
        """

        PyFunceble.facility.Logger.info(
            "Started to update setup.py.",
        )

        if self.branch == "dev":
            regexes = [
                (r'name=".*"', 'name="PyFunceble-dev"'),
                (r'"Development\sStatus\s::.*"', '"Development Status :: 4 - Beta"'),
            ]
        elif self.branch == "master":
            regexes = [
                (r'name=".*"', 'name="PyFunceble-dev"'),
                (
                    r'"Development\sStatus\s::.*"',
                    '"Development Status :: 5 - Production/Stable"',
                ),
            ]
        else:
            regexes = [
                (r'name=".*"', 'name="PyFunceble-dev"'),
                (
                    r'"Development\sStatus\s::.*"',
                    '"Development Status :: 3 - Alpha"',
                ),
            ]

        self.file_helper.set_path(
            os.path.join(PyFunceble.storage.CONFIG_DIRECTORY, "setup.py")
        )

        if not self.file_helper.exists():
            raise FileNotFoundError(self.file_helper.path)

        to_update = self.file_helper.read()

        for regex, replacement in regexes:
            to_update = self.regex_helper.set_regex(regex).replace_match(
                to_update, replacement, multiline=True
            )

        self.file_helper.write(to_update, overwrite=True)

        PyFunceble.facility.Logger.info(
            "Started to update setup.py.",
        )

        return self

    def update_version_file(self) -> "ProductionPrep":
        """
        Updates the version file.
        """

        PyFunceble.facility.Logger.info(
            "Started to update version file.",
        )

        if self.should_be_deprecated(self.previous_version):
            to_append = ".".join(
                self.version_utility.get_splitted(self.version_utility.local_version)[0]
            )

            if to_append not in self.version_file_content["deprecated"]:
                self.version_file_content["deprecated"].append(to_append)

        self.version_file_content[
            "current_version"
        ] = PyFunceble.storage.PROJECT_VERSION

        self.dict_helper.set_subject(self.version_file_content).to_yaml_file(
            self.VERSION_FILE_PATH
        )

        PyFunceble.facility.Logger.info(
            "Finished to update version file.",
        )

        return self

    def update_dir_structure_file(self) -> "ProductionPrep":
        """
        Updates the directory structure.
        """

        DirectoryStructureBackup().start()

        return self

    def start(self) -> "ProductionPrep":
        """
        Starts the production process.
        """

        return (
            self.update_setup_py()
            .update_code_urls()
            .update_code_format()
            .update_version_file()
        )



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the configuration comparision interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import copy
from typing import List, Optional

import PyFunceble.storage
from PyFunceble.helpers.dict import DictHelper
from PyFunceble.helpers.merge import Merge


class ConfigComparison:
    """
    Provides an interface for comparing 2 configuration.
    """

    DELETED_LINKS: List[str] = [
        "config",
        "dir_structure",
        "iana",
        "ipv4_reputation",
        "mariadb",
        "mysql",
        "psl",
        "repo",
        "requirements",
        "user_agents",
    ]
    DELETED_CORE: List[str] = [
        "dns_lookup_over_tcp",
        "generate_json",
        "header_printed",
        "iana_whois_server",
        "idna_conversion",
        "logs",
        "maximal_processes",
        "multiprocess_merging_mode",
        "multiprocess",
        "no_http_codes",
        "outputs",
        "shadow_file",
        "status",
        "store_whois_record",
        "unified",
    ]

    OLD_TO_NEW: dict = {
        "adblock": "cli_decoding.adblock",
        "aggressive": "cli_decoding.adblock_aggressive",
        "auto_continue": "cli_testing.autocontinue",
        "command": "cli_testing.ci.command",
        "command_before_end": "cli_testing.ci.end_command",
        "cooldown_time": "cli_testing.cooldown_time",
        "custom_ip": "cli_testing.hosts_ip",
        "days_between_inactive_db_clean": "cli_testing.days_between.db_clean",
        "days_between_db_retest": "cli_testing.days_between.db_retest",
        "db_type": "cli_testing.db_type",
        "debug": "debug.active",
        "dns_server": "dns.server",
        "filter": "cli_testing.file_filter",
        "generate_complements": "cli_testing.complements",
        "generate_hosts": "cli_testing.file_generation.hosts",
        "hierarchical_sorting": "cli_testing.sorting_mode.hierarchical",
        "inactive_database": "cli_testing.inactive_db",
        "less": "cli_testing.display_mode.less",
        "local": "cli_testing.local_network",
        "mining": "cli_testing.mining",
        "no_files": "cli_testing.file_generation.no_file",
        "plain_list_domain": "cli_testing.file_generation.plain",
        "print_dots": "cli_testing.display_mode.dots",
        "quiet": "cli_testing.display_mode.dots",
        "use_reputation_data": "lookup.reputation",
        "reputation": "lookup.reputation",
        "rpz": "cli_decoding.rpz",
        "show_execution_time": "cli_testing.display_mode.execution_time",
        "show_percentage": "cli_testing.display_mode.percentage",
        "simple": "cli_testing.display_mode.simple",
        "syntax": "cli_testing.testing_mode.syntax",
        "timeout": "lookup.timeout",
        "ci": "cli_testing.ci.active",
        "ci_autosave_commit": "cli_testing.ci.commit_message",
        "ci_autosave_final_commit": "cli_testing.ci.end_commit_message",
        "ci_autosave_minutes": "cli_testing.ci.max_exec_minutes",
        "ci_branch": "cli_testing.ci.branch",
        "ci_distribution_branch": "cli_testing.ci.distribution_branch",
        "whois_database": "cli_testing.whois_db",
        "wildcard": "cli_decoding.wildcard",
    }

    OLD_TO_NEW_NEGATE: dict = {
        "no_special": "lookup.special",
        "no_whois": "lookup.whois",
        "split": "cli_testing.file_generation.unified_results",
    }

    NEW_STATUS_CODES: dict = {
        "up": [102, 207, 208, 226, 429],
        "potentially_down": [451],
        "potentially_up": [
            308,
            403,
            418,
            421,
            422,
            423,
            424,
            426,
            428,
            431,
            506,
            507,
            508,
            510,
            511,
        ],
    }

    _local_config: dict = {}
    _upsteam_config: dict = {}

    dict_helper: DictHelper = DictHelper()

    def __init__(
        self,
        *,
        local_config: Optional[dict] = None,
        upstream_config: Optional[dict] = None,
    ) -> None:
        if local_config:
            self.local_config = local_config

        if upstream_config:
            self.upstream_config = upstream_config

    @property
    def local_config(self) -> dict:
        """
        Provides the current state of the :code:`_local_config`.
        """

        return self._local_config

    @local_config.setter
    def local_config(self, value: dict) -> None:
        """
        Sets the local configuration to work with.

        :raise TypeError:
            When :code:`value` is not a :py:class:`dict`.
        """

        if not isinstance(value, dict):
            raise TypeError(f"<value> should be {dict}, {type(value)} given.")

        self._local_config = copy.deepcopy(value)

    def set_local_config(self, value: dict) -> "ConfigComparison":
        """
        Sets the local configuration to work with.
        """

        self.local_config = value

        return self

    @property
    def upstream_config(self) -> dict:
        """
        Provides the current state of the :code:`_upstream_config`.
        """

        return self._upsteam_config

    @upstream_config.setter
    def upstream_config(self, value: dict) -> None:
        """
        Sets the upstram configuration to work with.

        :raise TypeError:
            When :code:`value` is not a :py:class:`dict`
        """

        if not isinstance(value, dict):
            raise TypeError(f"<value> should be {dict}, {type(value)} given.")

        self._upsteam_config = copy.deepcopy(value)

    def set_upstream_config(self, value: dict) -> "ConfigComparison":
        """
        Sets the upstram configuration to work with.
        """

        self.upstream_config = value

        return self

    def is_local_identical(self) -> bool:
        """
        Checks if the local configuration is identical to the upstream one.
        """

        # pylint: disable=too-many-boolean-expressions
        if (
            not self.dict_helper.set_subject(self.local_config).has_same_keys_as(
                self.upstream_config
            )
            or "user_agent" not in self.local_config
            or not isinstance(self.local_config["user_agent"], dict)
            or "active" in self.local_config["http_codes"]
            or "not_found_default" in self.local_config["http_codes"]
            or "self_managed" not in self.local_config["http_codes"]
            or "dns" not in self.local_config
            or "proxy" not in self.local_config
            or "follow_server_order" not in self.local_config["dns"]
            or "trust_server" not in self.local_config["dns"]
            or "collection" not in self.local_config
            or "collection" not in self.local_config["lookup"]
        ):
            return False

        for index in self.local_config:
            if index in self.DELETED_CORE:
                return False

        for index in self.local_config["links"]:
            if index in self.DELETED_LINKS:
                return False

        if "self_managed" in self.local_config["http_codes"] and not bool(
            self.local_config["http_codes"]["self_managed"]
        ):
            for index, values in self.local_config["http_codes"]["list"].items():
                if set(self.upstream_config["http_codes"]["list"][index]) != set(
                    values
                ):
                    return False

        return True

    def get_merged(self) -> dict:
        """
        Provides the merged configuration.
        """

        # pylint: disable=too-many-branches

        if self.is_local_identical():
            return self.local_config

        if not self.local_config:
            return self.upstream_config

        original_local = copy.deepcopy(self.local_config)
        original_upstream = copy.deepcopy(self.upstream_config)

        flatten_original = self.dict_helper.set_subject(original_local).flatten()
        flatten_upstream = self.dict_helper.set_subject(original_upstream).flatten()

        for key, value in self.OLD_TO_NEW.items():
            if key not in flatten_original:
                continue

            if value not in flatten_upstream:  # pragma: no cover ## Safety.
                raise RuntimeError(f"<value> ({value!r}) not found.")

            flatten_original[value] = original_local[key]

            del flatten_original[key]

        for key, value in self.OLD_TO_NEW_NEGATE.items():
            if key not in flatten_original:
                continue

            if value not in flatten_upstream:  # pragma: no cover ## Safety.0
                raise RuntimeError(f"<value> ({value!r}) not found.")

            flatten_original[value] = not original_local[key]

            del flatten_original[key]

        original_local = self.dict_helper.set_subject(flatten_original).unflatten()
        del flatten_original

        merged = Merge(original_local).into(original_upstream)

        if "dns_lookup_over_tcp" in merged and merged["dns_lookup_over_tcp"]:
            merged["dns"]["protocol"] = "TCP"

        for index in self.DELETED_CORE:
            if index in merged:
                del merged[index]

        for index in self.DELETED_LINKS:
            if index in merged["links"]:
                del merged["links"][index]

        if not bool(merged["http_codes"]["self_managed"]):
            for index, values in PyFunceble.storage.STD_HTTP_CODES.list.items():
                merged["http_codes"]["list"][index] = list(values)

        if merged["cli_testing"]["db_type"] == "json":
            merged["cli_testing"]["db_type"] = "csv"

        if merged["cli_testing"]["cooldown_time"] is None:
            merged["cli_testing"]["cooldown_time"] = self.upstream_config[
                "cli_testing"
            ]["cooldown_time"]

        if "user_agent" not in self.local_config or not isinstance(
            self.local_config["user_agent"], dict
        ):
            merged["user_agent"] = self.upstream_config["user_agent"]

        if "active" in merged["http_codes"]:
            del merged["http_codes"]["active"]

        if "not_found_default" in merged["http_codes"]:
            del merged["http_codes"]["not_found_default"]

        return merged



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the our HTTPS adapter.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import urllib.parse

import requests

import PyFunceble.facility
import PyFunceble.storage
from PyFunceble.query.requests.adapter.base import RequestAdapterBase


class RequestHTTPSAdapter(RequestAdapterBase):
    """
    Provides our HTTP adapter.
    """

    # pylint: disable=arguments-differ
    def send(self, request, **kwargs) -> requests.Response:
        """
        Overwrite the upstream :code:`send` method.

        We basically do the same. We only ensure that we request the IP from the chosen
        DNS record.

        :param request: The :class:`PreparedRequest <PreparedRequest>` being sent.
        :param stream: (optional) Whether to stream the request content.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or
            a :ref:`(connect timeout, read timeout) <timeouts>` tuple.
        :type timeout: float or tuple or urllib3 Timeout object
        :param verify: (optional) Either a boolean, in which case it controls whether
            we verify the server's TLS certificate, or a string, in which case it
            must be a path to a CA bundle to use
        :param cert: (optional) Any user-provided SSL certificate to be trusted.
        :param proxies: (optional) The proxies dictionary to apply to the request.
        :rtype: requests.Response

        .. versionchanged:: 4.1.0b16
            When a proxy is given, it is acceptable to have an unresolvable
            subject. The proxy should handle the situation and give us back a
            proper status or error.
        """

        kwargs["timeout"] = self.timeout

        parsed_url = urllib.parse.urlparse(request.url)
        hostname_ip = self.resolve(parsed_url.hostname)

        kwargs["proxies"] = self.fetch_proxy_from_pattern(parsed_url.hostname)

        PyFunceble.facility.Logger.debug("Parsed URL: %r", parsed_url)
        PyFunceble.facility.Logger.debug("Resolved IP: %r", hostname_ip)
        PyFunceble.facility.Logger.debug("KWARGS: %r", kwargs)
        PyFunceble.facility.Logger.debug(
            "Pool Manager: %r", self.poolmanager.connection_pool_kw
        )

        if hostname_ip or kwargs["proxies"]:
            if hostname_ip:
                request.url = request.url.replace(
                    f"{parsed_url.scheme}://{parsed_url.hostname}",
                    f"{parsed_url.scheme}://{hostname_ip}",
                )

            if parsed_url.scheme == "https":
                self.poolmanager.connection_pool_kw[
                    "server_hostname"
                ] = parsed_url.hostname
                self.poolmanager.connection_pool_kw[
                    "assert_hostname"
                ] = parsed_url.hostname

            # Ensure that the Hosts header is present. Otherwise, connection might
            # not work.
            request.headers["Host"] = parsed_url.hostname
        else:
            self.poolmanager.connection_pool_kw.pop(
                "server_hostname", PyFunceble.storage.NOT_RESOLVED_STD_HOSTNAME
            )
            self.poolmanager.connection_pool_kw.pop(
                "assert_hostname", PyFunceble.storage.NOT_RESOLVED_STD_HOSTNAME
            )

            self.poolmanager.connection_pool_kw.pop(
                "server_hostname", parsed_url.hostname
            )
            self.poolmanager.connection_pool_kw.pop(
                "assert_hostname", parsed_url.hostname
            )

            return self.fake_response()

        response = super().send(request, **kwargs)

        if hostname_ip:
            response.url = response.url.replace(hostname_ip, parsed_url.hostname)

        return response



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides a way to convert a given month to our unified format.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from typing import Any, Dict, List

from PyFunceble.query.whois.converter.base import ConverterBase


class Month2Unified(ConverterBase):
    """
    Converts the given month into our unified format.
    """

    MAP: Dict[str, List[str]] = {
        "jan": [str(1), "01", "jan", "january", "jan."],
        "feb": [str(2), "02", "feb", "february", "feb."],
        "mar": [str(3), "03", "mar", "march", "mar."],
        "apr": [str(4), "04", "apr", "april", "apr."],
        "may": [str(5), "05", "may"],
        "jun": [str(6), "06", "jun", "june", "jun."],
        "jul": [str(7), "07", "jul", "july", "jul."],
        "aug": [str(8), "08", "aug", "august", "aug."],
        "sep": [str(9), "09", "sep", "september", "sep.", "sept", "sept."],
        "oct": [str(10), "oct", "october", "oct."],
        "nov": [str(11), "nov", "november", "nov."],
        "dec": [str(12), "dec", "december", "dec."],
    }

    @ConverterBase.data_to_convert.setter
    def data_to_convert(self, value: Any) -> None:
        """
        Overrites the default behavior.

        :raise TypeError:
            When the given data to convert is not :py:class:`str`
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        # pylint: disable=no-member
        super(Month2Unified, self.__class__).data_to_convert.fset(self, value)

    @ConverterBase.ensure_data_to_convert_is_given
    def get_converted(self) -> str:
        """
        Provides the converted data (after conversion)

        .. warning::
            If no month is found, the given data is given as response.
        """

        for to_return, possibilities in self.MAP.items():
            if self.data_to_convert.lower() in possibilities:
                return to_return

        return self.data_to_convert



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the directory helpers.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import os
import shutil
from typing import List, Optional

from PyFunceble.helpers.list import ListHelper


class DirectoryHelper:
    """
    Simplify the directories manipulation.

    :param str dir_path the path to work with.
    """

    _path: Optional[str] = None

    def __init__(self, path: Optional[str] = None) -> None:
        if path is not None:
            self.path = path

    @property
    def path(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_path` attribute.
        """

        return self._path

    @path.setter
    def path(self, value: str) -> None:
        """
        Sets the directory path to work with.

        :param value:
            The path to set.

        :raise TypeError:
            When :code:`value` is not a :py:class:`str`.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        self._path = value

    def set_path(self, value: str) -> "DirectoryHelper":
        """
        Sets the directory path to work with.

        :param value:
            The path to set.
        """

        self.path = value

        return self

    @staticmethod
    def get_current(*, with_end_sep: bool = False) -> str:
        """
        Returns the current directory path.

        :param bool with_end_sep:
            Add a directory separator at the end.
        """

        if with_end_sep:
            return os.getcwd() + os.sep
        return os.getcwd()

    @property
    def realpath(self) -> str:
        """
        Returns the real path of the current path.
        """

        return os.path.realpath(self.path)

    def join_path(self, *args) -> str:
        """
        Joins the given arguments with the given path.
        """

        return os.path.join(self.path, *args)

    def exists(self) -> bool:
        """
        Checks if the given directory exists.
        """

        return os.path.isdir(self.path)

    def create(self) -> "DirectoryHelper":
        """
        Creates the given directory path.

        :return: The output of :code:`self.exists` after the directory creation.
        """

        if self.path and not self.exists():
            os.makedirs(self.path)

        return self

    def delete(self) -> "DirectoryHelper":
        """
        Deletes the given directory path.

        :return: :code:`not self.exists` after the directory deletion.
        """

        if self.exists():
            shutil.rmtree(self.path)

        return self

    def list_all_subdirectories(self) -> List[str]:
        """
        Provides the list of all subdirectories of the current path.
        """

        result = []

        if self.exists():
            for root, directories, _ in os.walk(self.path):
                for directory in directories:
                    result.append(os.path.join(root, directory))

        return ListHelper(result).remove_duplicates().sort().subject

    def list_all_files(self) -> List[str]:
        """
        Lists all files of the current path.
        """

        result = []

        if self.exists():
            for directory in self.list_all_subdirectories():
                for element in os.listdir(directory):
                    possible_element = os.path.join(directory, element)

                    if not os.path.isfile(possible_element):
                        continue

                    result.append(possible_element)

        return ListHelper(result).sort().subject



"""Delete the file table.

Revision ID: 83ada95132bf
Revises: 459a0d7b8f09
Create Date: 2020-12-07 12:49:48.797794

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql

# pylint: skip-file

# revision identifiers, used by Alembic.
revision = "83ada95132bf"
down_revision = "459a0d7b8f09"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "pyfunceble_status",
        "file_id",
        existing_type=mysql.INTEGER(display_width=11),
        nullable=True,
    )
    op.drop_constraint(
        "pyfunceble_status_ibfk_1", "pyfunceble_status", type_="foreignkey"
    )
    op.drop_index("path", table_name="pyfunceble_file")
    # op.drop_table('pyfunceble_file')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_foreign_key(
        "pyfunceble_status_ibfk_1",
        "pyfunceble_status",
        "pyfunceble_file",
        ["file_id"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    op.alter_column(
        "pyfunceble_status",
        "file_id",
        existing_type=mysql.INTEGER(display_width=11),
        nullable=False,
    )
    op.create_table(
        "pyfunceble_file",
        sa.Column(
            "id", mysql.INTEGER(display_width=11), autoincrement=True, nullable=False
        ),
        sa.Column("created", mysql.DATETIME(), nullable=False),
        sa.Column("modified", mysql.DATETIME(), nullable=True),
        sa.Column("path", mysql.TEXT(collation="utf8mb4_unicode_ci"), nullable=False),
        sa.Column(
            "test_completed",
            mysql.TINYINT(display_width=1),
            autoincrement=False,
            nullable=False,
        ),
        sa.CheckConstraint("`test_completed` in (0,1)", name="CONSTRAINT_1"),
        sa.PrimaryKeyConstraint("id"),
        mysql_collate="utf8mb4_unicode_ci",
        mysql_default_charset="utf8mb4",
        mysql_engine="InnoDB",
    )
    op.create_index("path", "pyfunceble_file", ["path"], unique=True)
    # ### end Alembic commands ###



# pylint: disable=invalid-name
"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import warnings
from typing import List

import PyFunceble.storage
from PyFunceble.checker.availability.domain import DomainAvailabilityChecker
from PyFunceble.checker.availability.domain_and_ip import DomainAndIPAvailabilityChecker
from PyFunceble.checker.availability.ip import IPAvailabilityChecker
from PyFunceble.checker.availability.status import AvailabilityCheckerStatus
from PyFunceble.checker.availability.url import URLAvailabilityChecker
from PyFunceble.checker.reputation.domain import DomainReputationChecker
from PyFunceble.checker.reputation.domain_and_ip import DomainAndIPReputationChecker
from PyFunceble.checker.reputation.ip import IPReputationChecker
from PyFunceble.checker.reputation.url import URLReputationChecker
from PyFunceble.checker.syntax.domain import DomainSyntaxChecker
from PyFunceble.checker.syntax.ip import IPSyntaxChecker
from PyFunceble.checker.syntax.ipv4 import IPv4SyntaxChecker
from PyFunceble.checker.syntax.ipv6 import IPv6SyntaxChecker
from PyFunceble.checker.syntax.second_lvl_domain import SecondLvlDomainSyntaxChecker
from PyFunceble.checker.syntax.subdomain import SubDomainSyntaxChecker
from PyFunceble.checker.syntax.url import URLSyntaxChecker
from PyFunceble.converter.subject2complements import Subject2Complements

__version__ = PyFunceble.storage.PROJECT_VERSION


def load_config(*args, **kwargs) -> None:
    """
    Placeholder before deletion.

    Since 4.0.0, you are not required to load the configuration before hand.
    If you still want too because you may want to use a special CLI related
    method, you can doing it so:

    ::

        import PyFunceble.facility

        PyFunceble.facility.ConfigLoader.start()
    """

    warnings.warn(
        "PyFunceble.load_config may be removed in the future."
        "As of today, because the configuration is not necessary (anymore), "
        "this method does nothing. Consider it a placeholder.I"
        "Please consider using PyFunceble.facility.ConfigLoader.start instead.",
        DeprecationWarning,
    )

    _, _ = args, kwargs


def test(subject: str, **kwargs) -> AvailabilityCheckerStatus:
    """
    Checks the avaialbility of the given subject assuming that it is a domain
    or an IP.

    .. warning::
        This method may be removed in the future.
        It is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import DomainAndIPAvailabilityChecker

            my_subject = "example.org"
            the_status = DomainAndIPAvailabilityChecker(
                my_subject
            ).get_status()

            # Get the status in dict format.
            print(the_status.to_dict())

            # Get the status in json format.
            print(the_status.to_json())

            # Check if it is available.
            print(f"{my_subject} is available ? {the_status.is_available()}")

    :param subject:
        The subject to work with.
    """

    warnings.warn(
        "PyFunceble.test may be removed in the future."
        "Please consider using PyFunceble.DomainAndIPAvailabilityChecker explicitly.",
        DeprecationWarning,
    )

    return DomainAndIPAvailabilityChecker(subject, **kwargs).get_status()


def url_test(subject: str, **kwargs) -> AvailabilityCheckerStatus:
    """
    Checks the availability of the given subject assuming that it is a URL.

    .. warning::
        This method may be removed in the future.
        It is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import URLAvailabilityChecker

            my_subject = "http://example.org"
            the_status = URLAvailabilityChecker(
                my_subject
            ).get_status()

            # Get the status in dict format.
            print(the_status.to_dict())

            # Get the status in json format.
            print(the_status.to_json())

            # Check if it is available.
            print(f"{my_subject} is available ? {the_status.is_available()}")

    :param subject:
        The subject to work with.
    """

    warnings.warn(
        "PyFunceble.test may be removed in the future."
        "Please consider using PyFunceble.URLAvailabilityChecker explicitly.",
        DeprecationWarning,
    )

    return URLAvailabilityChecker(subject, **kwargs).get_status()


def is_second_level_domain(subject: str, **kwargs) -> bool:
    """
    Checks if the given subject is a syntactically valid second level domain.

    .. warning::
        This method was added for retrocompatibility.
        It may be removed in the future and is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import SecondLvlDomainSyntaxChecker

            my_subject = "example.org"
            the_status = SecondLvlDomainSyntaxChecker(
                my_subject
            ).get_status()

            # Get the status in dict format.
            print(the_status.to_dict())

            # Get the status in json format.
            print(the_status.to_json())

            # Check if it is a second level domain.
            print(f"{my_subject} is 2nd level domain ? {the_status.is_valid()}")

    :param subject:
        The subject to work with.
    """

    warnings.warn(
        "PyFunceble.is_second_level_domain was added for retro compatibility. "
        "It may be removed in the future."
        "Please consider using PyFunceble.SecondLvlDomainSyntaxChecker explicitly.",
    )

    return SecondLvlDomainSyntaxChecker(subject, **kwargs).is_valid()


def is_domain(subject: str, **kwargs) -> bool:
    """
    Checks if the given subject is a syntactically valid second level domain
    or subdomain.

    .. warning::
        This method may be removed in the future.
        It is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import DomainSyntaxChecker

            my_subject = "example.org"
            the_status = DomainSyntaxChecker(
                my_subject
            ).get_status()

            # Get the status in dict format.
            print(the_status.to_dict())

            # Get the status in json format.
            print(the_status.to_json())

            # Check if it is a domain (2nd level or subdomain).
            print(f"{my_subject} is domain ? {the_status.is_valid()}")

    :param subject:
        The subject to work with.
    """

    warnings.warn(
        "PyFunceble.is_domain may be removed in the future."
        "Please consider using PyFunceble.DomainSyntaxChecker explicitly.",
        DeprecationWarning,
    )

    return DomainSyntaxChecker(subject, **kwargs).is_valid()


def is_subdomain(subject: str, **kwargs) -> bool:
    """
    Checks if the given subject is a syntactically valid subdomain.

    .. warning::
        This method may be removed in the future.
        It is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import SubDomainSyntaxChecker

            my_subject = "hello.example.org"
            the_status = SubDomainSyntaxChecker(
                my_subject
            ).get_status()

            # Get the status in dict format.
            print(the_status.to_dict())

            # Get the status in json format.
            print(the_status.to_json())

            # Check if it is a subdomain.
            print(f"{my_subject} is subdomain ? {the_status.is_valid()}")

    :param subject:
        The subject to work with.
    """

    warnings.warn(
        "PyFunceble.is_subdomain may be removed in the future."
        "Please consider using PyFunceble.SubDomainSyntaxChecker explicitly.",
        DeprecationWarning,
    )

    return SubDomainSyntaxChecker(subject, **kwargs).is_valid()


def is_ip(subject: str, **kwargs) -> bool:
    """
    Checks if the given subject is a syntactically valid IP range.

    .. warning::
        This method may be removed in the future.
        It is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import IPSyntaxChecker

            my_subject = "192.168.0.0"
            the_status = IPSyntaxChecker(
                my_subject
            ).get_status()

            # Get the status in dict format.
            print(the_status.to_dict())

            # Get the status in json format.
            print(the_status.to_json())

            # Check if it is an IP (v4 or v6).
            print(f"{my_subject} is IP ? {the_status.is_valid()}")

    :param subject:
        The subject to work with.
    """

    warnings.warn(
        "PyFunceble.is_ip may be removed in the future."
        "Please consider using PyFunceble.IPSyntaxChecker explicitly.",
        DeprecationWarning,
    )

    return IPSyntaxChecker(subject, **kwargs).is_valid()


def is_ipv4(subject: str, **kwargs) -> bool:
    """
    Checks if the given subject is a syntactically valid IPv4.

    .. warning::
        This method may be removed in the future.
        It is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import IPSyntaxChecker

            my_subject = "192.168.0.0"
            the_status = IPSyntaxChecker(
                my_subject
            ).get_status()

            # Get the status in dict format.
            print(the_status.to_dict())

            # Get the status in json format.
            print(the_status.to_json())

            # Check if it is an IPv4.
            print(f"{my_subject} is IPv4 ? {the_status.is_valid_v4()}")

    :param subject:
        The subject to work with.
    """

    warnings.warn(
        "PyFunceble.is_ipv4 be removed in the future."
        "Please consider using PyFunceble.IPSyntaxChecker explicitly.",
        DeprecationWarning,
    )

    return IPSyntaxChecker(subject, **kwargs).is_valid_v4()


def is_ipv6(subject: str, **kwargs) -> bool:
    """
    Checks if the given subject is a syntactically valid IPv6.

    .. warning::
        This method may be removed in the future.
        It is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import IPSyntaxChecker

            my_subject = "192.168.0.0"
            the_status = IPSyntaxChecker(
                my_subject
            ).get_status()

            # Get the status in dict format.
            print(the_status.to_dict())

            # Get the status in json format.
            print(the_status.to_json())

            # Check if it is an IPv6.
            print(f"{my_subject} is IPv6 ? {the_status.is_valid_v6()}")

    :param subject:
        The subject to work with.
    """

    warnings.warn(
        "PyFunceble.is_ipv6 be removed in the future."
        "Please consider using PyFunceble.IPSyntaxChecker explicitly.",
        DeprecationWarning,
    )

    return IPSyntaxChecker(subject, **kwargs).is_valid_v6()


def is_ip_range(subject: str, **kwargs) -> bool:
    """
    Checks if the given subject is a syntactically valid IP range.

    .. warning::
        This method may be removed in the future.
        It is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import IPSyntaxChecker

            my_subject = "192.168.0.0"
            the_status = IPSyntaxChecker(
                my_subject
            ).get_status()

            # Get the status in dict format.
            print(the_status.to_dict())

            # Get the status in json format.
            print(the_status.to_json())

            # Check if it is an IP range (v4 or v6).
            print(f"{my_subject} is IP range ? {the_status.is_valid_range()}")

    :param subject:
        The subject to work with.
    """

    warnings.warn(
        "PyFunceble.is_ip_range may be removed in the future."
        "Please consider using PyFunceble.IPSyntaxChecker explicitly.",
        DeprecationWarning,
    )

    return IPSyntaxChecker(subject, **kwargs).is_valid_range()


def is_ipv4_range(subject: str, **kwargs) -> bool:
    """
    Checks if the given subject is a syntactically valid IPv4 range.

    .. warning::
        This method may be removed in the future.
        It is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import IPSyntaxChecker

            my_subject = "192.168.0.0"
            the_status = IPSyntaxChecker(
                my_subject
            ).get_status()

            # Get the status in dict format.
            print(the_status.to_dict())

            # Get the status in json format.
            print(the_status.to_json())

            # Check if it is IPv4 range.
            print(f"{my_subject} is IPv4 range ? {the_status.is_valid_v4_range()}")

    :param subject:
        The subject to work with.
    """

    warnings.warn(
        "PyFunceble.is_ipv4_range may be removed in the future."
        "Please consider using PyFunceble.IPSyntaxChecker explicitly.",
        DeprecationWarning,
    )

    return IPSyntaxChecker(subject, **kwargs).is_valid_v4_range()


def is_ipv6_range(subject: str, **kwargs) -> bool:
    """
    Checks if the given subject is a syntactically valid IPv6 range.

    .. warning::
        This method may be removed in the future.
        It is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import IPSyntaxChecker

            my_subject = "::1"
            the_status = IPSyntaxChecker(
                my_subject
            ).get_status()

            # Get the status in dict format.
            print(the_status.to_dict())

            # Get the status in json format.
            print(the_status.to_json())

            # Check if it is IPv6 range.
            print(f"{my_subject} is IPv6 range ? {the_status.is_valid_v6_range()}")

    :param subject:
        The subject to work with.
    """

    warnings.warn(
        "PyFunceble.is_ipv6_range may be removed in the future."
        "Please consider using PyFunceble.IPSyntaxChecker explicitly.",
        DeprecationWarning,
    )

    return IPSyntaxChecker(subject, **kwargs).is_valid_v6_range()


def is_url(subject: str, **kwargs) -> bool:
    """
    Checks if the given subject is syntactically a valid URL.

    .. warning::
        This method may be removed in the future.
        It is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import DomainReputationChecker

            my_subject = "https://example.org"
            the_status = URLSyntaxChecker(
                my_subject
            ).get_status()

            # Get the status in dict format.
            print(the_status.to_dict())

            # Get the status in json format.
            print(the_status.to_json())

            # Check if it is a URL.
            print(f"{my_subject} is URL ? {the_status.is_valid()}")

    :parma subject:
        The subject to check.
    """

    warnings.warn(
        "PyFunceble.is_url may be removed in the future."
        "Please consider using PyFunceble.URLSyntaxChecker explicitly.",
        DeprecationWarning,
    )

    # pylint: disable=no-member
    return URLSyntaxChecker(subject, **kwargs).get_status().is_valid()


def is_domain_malicious(subject: str, **kwargs) -> bool:
    """
    Checks if the given domain is malicious.

    .. warning::
        This method may be removed in the future.
        It is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import DomainReputationChecker

            my_subject = "example.org"
            the_status = DomainReputationChecker(
                my_subject
            ).get_status()

            # Get the status in dict format.
            print(the_status.to_dict())

            # Get the status in json format.
            print(the_status.to_json())

            # Check if it is malicious.
            print(f"{my_subject} is Malicious ? {the_status.is_malicious()}")

    :param subject:
        The subject to work with.
    """

    warnings.warn(
        "PyFunceble.is_domain_malicious may be removed in the future."
        "Please consider using PyFunceble.DomainReputationChecker explicitly.",
        DeprecationWarning,
    )

    # pylint: disable=no-member
    return DomainReputationChecker(subject, **kwargs).get_status().is_malicious()


def is_ipv4_malicious(subject: str, **kwargs) -> bool:
    """
    Checks if the given IPv4 is malicious.

    .. warning::
        This method may be removed in the future.
        It is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import IPReputationChecker

            my_subject = "192.168.0.1"
            the_status = IPReputationChecker(
                my_subject
            ).get_status()

            # Get the status in dict format.
            print(the_status.to_dict())

            # Get the status in json format.
            print(the_status.to_json())

            # Check if it is malicious.
            print(f"{my_subject} is Malicious ? {the_status.is_malicious()}")

    :param subject:
        The subject to work with.
    """

    warnings.warn(
        "PyFunceble.is_ipv4_malicious may be removed in the future."
        "Please consider using PyFunceble.IPReputationChecker explicitly.",
        DeprecationWarning,
    )

    # pylint: disable=no-member
    return IPReputationChecker(subject, **kwargs).get_status().is_malicious()


def is_url_malicious(subject: str, **kwargs) -> bool:
    """
    Checks if the given URL is malicious.

    .. warning::
        This method may be removed in the future.
        It is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import URLReputationChecker

            my_subject = "https://example.org"
            the_status = URLReputationChecker(
                my_subject
            ).get_status()

            # Get the status in dict format.
            print(the_status.to_dict())

            # Get the status in json format.
            print(the_status.to_json())

            # Check if it is malicious.
            print(f"{my_subject} is Malicious ? {the_status.is_malicious()}")

    :param subject:
        The subject to work with.
    """

    warnings.warn(
        "PyFunceble.is_url_malicious may be removed in the future."
        "Please consider using PyFunceble.URLReputationChecker explicitly.",
        DeprecationWarning,
    )

    # pylint: disable=no-member
    return URLReputationChecker(subject, **kwargs).get_status().is_malicious()


def get_complements(subject: str, include_given: bool = False) -> List[str]:
    """
    Provides the complements of a given subject.

    A complement is a for example :code:`example.org` if :code:`www.example.org`
    is given and vice-versa.

    .. warning::
        This method may be removed in the future.
        It is still available for convenience.

        Please consider the following alternative example:

        ::

            from PyFunceble import Subject2Complements

            my_subject = "example.org"
            complements = Subject2Complements(
                my_subject
            ).get_converted(include_given=True)

    :param subject:
        The subject to work with.

    :param include_given:
        Include the given subject in the result.
    """

    warnings.warn(
        "PyFunceble.get_complements may be removed in the future."
        "Please consider using PyFunceble.Subject2Complements explicitly",
        DeprecationWarning,
    )

    return Subject2Complements(subject, include_given=include_given).get_converted()



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides a base for the manipulation of JSON files.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import copy
import functools
import os
from typing import Dict, Optional, Union

from PyFunceble.cli.filesystem.dir_base import FilesystemDirBase
from PyFunceble.helpers.dict import DictHelper
from PyFunceble.helpers.file import FileHelper


class FilesystemJSONBase(FilesystemDirBase):
    """
    A base interface for the manipulation of JSON files.
    """

    dataset: Dict[str, int] = {}
    source_file_path: Optional[str] = None

    SOURCE_FILE: Optional[str] = None
    STD_DATASET: Union[dict, list] = {}

    def update_source_file_path_beforehand(func):  # pylint: disable=no-self-argument
        """
        Updates the source file before launching the decorated method.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            self.source_file_path = os.path.join(
                self.get_output_basedir(), self.SOURCE_FILE
            )

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    def fetch_dataset_beforehand(func):  # pylint: disable=no-self-argument
        """
        Updates the dataset to work with before launching the decorated method.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            self.fetch_dataset()

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    def save_dataset_afterwards(func):  # pylint: disable=no-self-argument
        """
        Saves the dataset after launching the decorated method.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            result = func(self, *args, **kwargs)  # pylint: disable=not-callable

            self.save_dataset()

            return result

        return wrapper

    @update_source_file_path_beforehand
    def fetch_dataset(self) -> "FilesystemJSONBase":
        """
        Fetch the dataset from the source file.
        """

        file_helper = FileHelper(self.source_file_path)

        if file_helper.exists():
            self.dataset = DictHelper().from_json_file(file_helper.path)
        else:
            self.dataset = copy.deepcopy(self.STD_DATASET)

        return self

    def save_dataset(self) -> "FilesystemJSONBase":
        """
        Saves the current dataset into it's final destination.
        """

        DictHelper(self.dataset).to_json_file(self.source_file_path)

        return self



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides a way to provides the nameserver to use.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import functools
from typing import List, Optional, Union

import dns.resolver

import PyFunceble.facility
import PyFunceble.storage
from PyFunceble.query.dns.nameserver import Nameservers


class Resolver:
    """
    Provides the right resolver.

    :param nameservers:
        The list of nameservers to communicate with.
    """

    STD_TIMEOUT: float = 5.0

    timeout: float = 3.0

    nameservers: Nameservers = Nameservers()
    internal_resolver: Optional[dns.resolver.Resolver] = None

    def __init__(
        self, nameservers: Optional[List[str]] = None, timeout: Optional[float] = None
    ) -> None:
        if nameservers is not None:
            self.set_nameservers(nameservers)
        else:
            self.nameservers.guess_and_set_nameservers()

        if timeout is not None:
            self.set_timeout(timeout)
        else:
            self.guess_and_set_timeout()

    def configure_resolver(func):  # pylint: disable=no-self-argument
        """
        Configures the resolvers after calling the decorated method.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            result = func(self, *args, **kwargs)  # pylint: disable=not-callable

            self.internal_resolver.lifetime = self.timeout + 2.0
            self.internal_resolver.timeout = self.timeout
            self.internal_resolver.nameservers = self.nameservers.get_nameservers()
            self.internal_resolver.nameserver_ports = (
                self.nameservers.get_nameserver_ports()
            )

            return result

        return wrapper

    def set_nameservers(self, value: List[str]) -> "Resolver":
        """
        Sets the given nameserver.
        """

        self.nameservers.set_nameservers(value)

    def set_timeout(self, value: Union[float, int]) -> "Resolver":
        """
        Sets the timeout of a query.
        """

        if not isinstance(value, (float, int)):
            raise TypeError(f"<value> should be {float}, {type(value)} given.")

        self.timeout = float(value)

        return self

    def get_nameservers(self) -> Optional[List[str]]:
        """
        Provides the currently set list of nameserver.
        """

        return self.nameservers.get_nameservers()

    def get_nameserver_ports(self) -> Optional[dict]:
        """
        Provides the currently set list of nameserver ports.
        """

        return self.nameservers.get_nameserver_ports()

    def get_timeout(self) -> Optional[float]:
        """
        Provides the currently set query timeout.
        """

        return self.timeout

    def guess_and_set_timeout(self) -> "Resolver":
        """
        Tries to guess the the timeout from the configuration.
        """

        if PyFunceble.facility.ConfigLoader.is_already_loaded():
            self.set_timeout(float(PyFunceble.storage.CONFIGURATION.lookup.timeout))
        else:
            self.set_timeout(self.STD_TIMEOUT)

        return self

    def guess_all_settings(
        self,
    ) -> "Resolver":  # pragma: no cover ## Method themselves are more important
        """
        Try to guess all settings.
        """

        to_ignore = ["guess_all_settings"]

        for method in dir(self):
            if method in to_ignore or not method.startswith("guess_"):
                continue

            getattr(self, method)()

        return self

    @configure_resolver
    def get_resolver(self) -> dns.resolver.Resolver:
        """
        Provides the resolver to work with.
        """

        if self.internal_resolver:
            return self.internal_resolver

        if self.nameservers.get_nameservers():
            self.internal_resolver = dns.resolver.Resolver(configure=False)
        else:  # pragma: no cover ## I don't want to play with the default resolver.
            self.internal_resolver = dns.resolver.Resolver()

        return self.internal_resolver



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the base of all our workers.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import multiprocessing
import multiprocessing.connection
import queue
import time
import traceback
from datetime import datetime, timedelta
from typing import Any, List, Optional, Tuple

import PyFunceble.cli.facility
import PyFunceble.cli.factory
import PyFunceble.facility
import PyFunceble.sessions
from PyFunceble.cli.continuous_integration.base import ContinuousIntegrationBase


class WorkerBase(multiprocessing.Process):
    """
    Provides the base of all our workers.

    :param input_queue:
        The input queue to read.
    :param output_queue:
        The output queue to write.
    """

    STD_NAME: str = "pyfunceble_base_worker"

    MINING_WAIT_TIME: int = 60
    BREAKOFF: float = 2.0

    input_queue: Optional[queue.Queue] = None
    output_queue: Optional[queue.Queue] = None

    continuous_integration: ContinuousIntegrationBase = None

    global_exit_event: Optional[multiprocessing.Event] = None
    exit_it: Optional[multiprocessing.Event] = None

    send_stop_message: Optional[bool] = None
    accept_waiting_delay: Optional[bool] = None

    concurrent_worker_names: Optional[List[str]] = None
    db_session: Optional[PyFunceble.cli.factory.db_session] = None

    _parent_connection: Optional[multiprocessing.connection.Connection] = None
    _child_connection: Optional[multiprocessing.connection.Connection] = None
    _exception: Optional[multiprocessing.Pipe] = None

    _params: Optional[dict] = {}

    def __init__(
        self,
        input_queue: Optional[queue.Queue],
        output_queue: Optional[queue.Queue] = None,
        global_exit_event: Optional[multiprocessing.Event] = None,
        *,
        name: Optional[str] = None,
        daemon: Optional[bool] = None,
        continuous_integration: Optional[ContinuousIntegrationBase] = None,
        configuration: Optional[dict] = None,
    ) -> None:
        self.configuration = self._params["configuration"] = configuration
        self.input_queue = self._params["input_queue"] = input_queue
        self.output_queue = self._params["output_queue"] = output_queue

        self.continuous_integration = self._params[
            "continuous_integration"
        ] = continuous_integration

        self.global_exit_event = self._params["global_exit_event"] = global_exit_event
        self.exit_it = multiprocessing.Event()

        self._parent_connection, self._child_connection = multiprocessing.Pipe()
        self._exception = None

        self.send_stop_message = True
        self.accept_waiting_delay = True
        self.concurrent_worker_names = []

        try:
            self.db_session = (
                PyFunceble.cli.factory.DBSession.get_db_session().get_new_session()()
            )
        except TypeError:
            self.db_session = None

        super().__init__(name=name, daemon=daemon)

        self.__post_init__()

    def __del__(self) -> None:
        if self.db_session is not None:
            self.db_session.close()

    def __post_init__(self) -> None:
        """
        A method which will be executed after the :code:`__init__` method.
        """

    @property
    def exception(self):
        """
        Provides the exception of the current worker.
        """

        if self._parent_connection.poll():
            self._exception = self._parent_connection.recv()

        return self._exception

    def add_to_input_queue(
        self,
        data: Any,
        *,
        worker_name: Optional[str] = None,
        destination_worker: Optional[str] = None,
    ) -> "WorkerBase":
        """
        Adds the given data to the current queue.

        :param data:
            The data to add into the queue.
        :param destination_worker:
            The name of the worker which is supposed to read the message.
        """

        if worker_name:
            to_send = (worker_name, destination_worker, data)
        else:
            to_send = (self.name, destination_worker, data)

        self.input_queue.put(to_send)

        PyFunceble.facility.Logger.debug("Added to the (input) queue: %r", data)

    def add_to_output_queue(
        self,
        data: Any,
        *,
        worker_name: Optional[str] = None,
        destination_worker: Optional[str] = None,
    ) -> "WorkerBase":
        """
        Adds the given data to the output queue queue.

        :param data:
            The data to add into the queue.
        """

        if worker_name:
            to_send = (worker_name, destination_worker, data)
        else:
            to_send = (self.name, destination_worker, data)

        if self.output_queue is not None:
            for output_queue in self.output_queue:
                output_queue.put(to_send)

        PyFunceble.facility.Logger.debug("Added to the (output) queue: %r", data)

        return self

    def target(self, consumed: Any) -> Optional[Tuple[Any, ...]]:
        """
        This the target that is run to process something.
        This method should return a result which will pu send to the output
        queue.
        """

        raise NotImplementedError()

    def share_waiting_message(
        self,
        *,
        overall: bool = False,
        ignore_current_worker: bool = False,
        apply_breakoff: bool = False,
    ) -> "WorkerBase":
        """
        Shares a waiting message to both input and output queues.
        The idea is to keep our worker awake when necessary. This functionality
        is really important when running the mining worker because sometime
        the flow is so long, that we may need some time.

        :param overall:
            Shares the message multiple time to specific destination in our
            current stack.
        :param ignore_current_worker:
            Doesn't share the message with ourself.
        :param apply_breakoff:
            Activates the breakoff (sleep) before starting to share the message.
        """

        if overall:
            for worker_name in self.concurrent_worker_names:
                if apply_breakoff:
                    time.sleep(self.BREAKOFF)

                if not ignore_current_worker:
                    # Give the whole stack a bit more of time.
                    self.add_to_input_queue("wait", destination_worker=worker_name)

                self.add_to_output_queue("wait")
        else:
            if apply_breakoff:
                time.sleep(self.BREAKOFF)

            if not ignore_current_worker:
                self.add_to_input_queue("wait", destination_worker=self.name)

            self.add_to_output_queue("wait")

    def run(self) -> None:  # pylint: disable=too-many-statements
        def break_now() -> bool:
            """
            Checks if it is time to make a break.
            """

            if not wait_for_stop or not self.accept_waiting_delay:
                return True

            return datetime.utcnow() > break_time

        if self.configuration is not None:
            PyFunceble.facility.ConfigLoader.set_custom_config(self.configuration)

        if multiprocessing.get_start_method() != "fork":
            PyFunceble.facility.ConfigLoader.start()
            PyFunceble.cli.facility.CredentialLoader.start()
            PyFunceble.cli.factory.DBSession.init_db_sessions()

        # Be sure that all settings are loaded proprely!!
        PyFunceble.factory.Requester = PyFunceble.factory.requester()

        wait_for_stop = (
            bool(PyFunceble.storage.CONFIGURATION.cli_testing.mining) is True
        )
        break_time = datetime.utcnow() + timedelta(seconds=self.MINING_WAIT_TIME)

        try:  # pylint: disable=too-many-nested-blocks
            while True:
                if self.global_exit_event.is_set():
                    PyFunceble.facility.Logger.info(
                        "Got global exit event. Stopping worker."
                    )

                    break

                if self.exit_it.is_set():
                    PyFunceble.facility.Logger.info("Got exit event. Stopping worker.")
                    break

                if (
                    self.continuous_integration
                    and self.continuous_integration.is_time_exceeded()
                ):
                    PyFunceble.facility.Logger.info(
                        "CI time exceeded. Stopping worker."
                    )

                    if break_now():
                        self.exit_it.set()
                        continue

                    continue

                try:
                    worker_name, destination_worker, consumed = self.input_queue.get()
                except EOFError:
                    PyFunceble.facility.Logger.info("Got EOFError. Stopping worker.")
                    self.global_exit_event.set()
                    break

                if destination_worker and destination_worker != self.name:
                    self.add_to_input_queue(
                        consumed,
                        worker_name=worker_name,
                        destination_worker=destination_worker,
                    )
                    continue

                PyFunceble.facility.Logger.info(
                    "Got (from %r): %r",
                    worker_name,
                    consumed,
                )

                if consumed == "stop":
                    if break_now():
                        PyFunceble.facility.Logger.info(
                            "Got stop message from %r. Applying.",
                            worker_name,
                        )

                        self.exit_it.set()
                        continue

                    self.share_waiting_message(apply_breakoff=wait_for_stop)
                    continue

                if consumed == "wait":
                    if not wait_for_stop:
                        continue

                    if break_now():
                        PyFunceble.facility.Logger.debug(
                            "Waited sufficiently. Stopping current worker."
                        )

                        # The real (break) action shouldn't be handled here.
                        self.exit_it.set()
                        continue

                    PyFunceble.facility.Logger.debug(
                        "We need to wait a bit more. Continue waiting."
                    )

                    self.share_waiting_message(apply_breakoff=wait_for_stop)
                    continue

                result = self.target(consumed)

                if result is not None:
                    self.add_to_output_queue(result)

                    PyFunceble.facility.Logger.info(
                        "Produced: %r",
                        result,
                    )

                break_time = datetime.utcnow() + timedelta(
                    seconds=self.MINING_WAIT_TIME
                )

        except Exception as exception:  # pylint: disable=broad-except
            PyFunceble.facility.Logger.critical(
                "Error while running target", exc_info=True
            )
            trace = traceback.format_exc()
            self._child_connection.send((exception, trace))

            self.exit_it.set()
            raise exception

    def terminate(self) -> None:
        """
        Terminate our worker.
        """

        self.exit_it.set()

        if self.input_queue:
            # Necessary to avoid blocking.
            self.add_to_input_queue("stop", destination_worker=self.name)



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides our tester worker. This is the description of a single tester worker.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import time
from typing import Any, Optional, Tuple

import PyFunceble.cli.utils.testing
import PyFunceble.facility
import PyFunceble.factory
from PyFunceble.checker.availability.domain_and_ip import DomainAndIPAvailabilityChecker
from PyFunceble.checker.availability.url import URLAvailabilityChecker
from PyFunceble.checker.base import CheckerBase
from PyFunceble.checker.reputation.domain_and_ip import DomainAndIPReputationChecker
from PyFunceble.checker.reputation.url import URLReputationChecker
from PyFunceble.checker.syntax.domain import DomainSyntaxChecker
from PyFunceble.checker.syntax.ip import IPSyntaxChecker
from PyFunceble.checker.syntax.url import URLSyntaxChecker
from PyFunceble.cli.processes.workers.base import WorkerBase
from PyFunceble.cli.utils.stdout import print_single_line
from PyFunceble.dataset.autocontinue.base import ContinueDatasetBase
from PyFunceble.dataset.autocontinue.csv import CSVContinueDataset
from PyFunceble.dataset.inactive.base import InactiveDatasetBase
from PyFunceble.helpers.regex import RegexHelper


class TesterWorker(WorkerBase):
    """
    Provides our tester worker. The objective of this worker is to provides
    a single worker (or process if you prefer) which will be used to handle
    the tests.
    """

    STD_NAME: str = "pyfunceble_tester_worker"

    continue_dataset: Optional[ContinueDatasetBase] = None
    inactive_dataset: Optional[InactiveDatasetBase] = None
    testing_object: Optional[CheckerBase] = None

    known_testing_objects: dict = {}
    initiated_testing_objects: dict = {}

    def __post_init__(self) -> None:
        self.continue_dataset = (
            PyFunceble.cli.utils.testing.get_continue_databaset_object(
                db_session=self.db_session
            )
        )
        self.inactive_dataset = (
            PyFunceble.cli.utils.testing.get_inactive_dataset_object(
                db_session=self.db_session
            )
        )

        self.known_testing_objects = {
            "SYNTAX": {"domain": DomainSyntaxChecker, "url": URLSyntaxChecker},
            "AVAILABILITY": {
                "domain": DomainAndIPAvailabilityChecker,
                "url": URLAvailabilityChecker,
            },
            "REPUTATION": {
                "domain": DomainAndIPReputationChecker,
                "url": URLReputationChecker,
            },
        }

        self.initiated_testing_objects = {
            "SYNTAX": {"domain": None, "url": None},
            "AVAILABILITY": {
                "domain": None,
                "url": None,
            },
            "REPUTATION": {
                "domain": None,
                "url": None,
            },
        }

        return super().__post_init__()

    @staticmethod
    def should_be_ignored(subject: str) -> bool:
        """
        Checks if the given subject should be ignored.
        """

        # pylint: disable=line-too-long
        regex_ignore = r"localhost$|localdomain$|local$|broadcasthost$|0\.0\.0\.0$|allhosts$|allnodes$|allrouters$|localnet$|loopback$|mcastprefix$|ip6-mcastprefix$|ip6-localhost$|ip6-loopback$|ip6-allnodes$|ip6-allrouters$|ip6-localnet$"

        if RegexHelper(regex_ignore).match(subject, return_match=False):
            PyFunceble.facility.Logger.info(
                "Ignoring %r because it is in our default regex.", subject
            )
            return True

        if (
            not PyFunceble.storage.CONFIGURATION.cli_testing.local_network
            and IPSyntaxChecker(subject).is_reserved()
        ):
            PyFunceble.facility.Logger.info(
                "Ignoring %r because it is a reserved IP and we are not testing "
                "for/in a local network.",
                subject,
            )
            return True

        if bool(
            PyFunceble.storage.CONFIGURATION.cli_testing.file_filter
        ) and not RegexHelper(
            PyFunceble.storage.CONFIGURATION.cli_testing.file_filter
        ).match(
            subject, return_match=False
        ):
            PyFunceble.facility.Logger.info(
                "Ignoring %r because it does not match the filter to look for.",
                subject,
            )
            return True

        PyFunceble.facility.Logger.info(
            "Allowed to test %r.",
            subject,
        )

        return False

    def _init_testing_object(
        self, subject_type: str, checker_type: str
    ) -> Optional[CheckerBase]:
        """
        Provides the object to use for testing.

        :raise ValueError:
            When the given subject type is unknown.
        """

        if checker_type not in self.known_testing_objects:
            raise ValueError(f"<testing_mode> ({checker_type!r}) is unknown.")

        if subject_type not in self.known_testing_objects[checker_type]:
            raise ValueError(f"<subject_type> ({subject_type!r}) is unknown.")

        if not self.initiated_testing_objects[checker_type][subject_type]:
            self.initiated_testing_objects[checker_type][
                subject_type
            ] = self.known_testing_objects[checker_type][subject_type](
                db_session=self.db_session
            ).set_do_syntax_check_first(
                # We want to always check the syntax first (ONLY UNDER THE CLI)
                not bool(PyFunceble.storage.CONFIGURATION.cli_testing.local_network)
            )

        self.testing_object = self.initiated_testing_objects[checker_type][subject_type]

        return self.testing_object

    def target(self, consumed: dict) -> Optional[Tuple[Any, ...]]:
        """
        This the target that is run to process something.
        This method should return a result which will pu send to the output
        queue.
        """

        if (
            PyFunceble.factory.Requester.session
            and "Connection" not in PyFunceble.factory.Requester.session.headers
        ):
            # Just close the connection immediately. This prevent potential infinite
            # streams.
            PyFunceble.factory.Requester.session.headers["Connection"] = "close"

        if not isinstance(consumed, dict):
            PyFunceble.facility.Logger.debug(
                "Skipping latest dataset because consumed data was not "
                "a dictionnary."
            )
            return None

        # Just for human brain.
        test_dataset = consumed

        if self.should_be_ignored(test_dataset["idna_subject"]):
            # X means that it was ignored because of our core ignore procedure.
            print_single_line("X")
            return None

        if PyFunceble.storage.CONFIGURATION.cli_testing.cooldown_time > 0:
            PyFunceble.facility.Logger.info(
                "Sleeping: %rs for our own safety :-)",
                PyFunceble.storage.CONFIGURATION.cli_testing.cooldown_time,
            )
            # Apply cooldowntime.
            time.sleep(PyFunceble.storage.CONFIGURATION.cli_testing.cooldown_time)
            PyFunceble.facility.Logger.info(
                "Slept: %rs for our own safety :-)",
                PyFunceble.storage.CONFIGURATION.cli_testing.cooldown_time,
            )

        if test_dataset["type"] != "single":
            if test_dataset["output_dir"] and "from_preload" not in test_dataset:
                if isinstance(self.continue_dataset, CSVContinueDataset):
                    self.continue_dataset.set_base_directory(test_dataset["output_dir"])

                if self.continue_dataset.exists(test_dataset):
                    # A means that it was ignored because of the continue
                    # logic.

                    PyFunceble.facility.Logger.info(
                        "Ignoring %r because it was already tested previously "
                        "(continue).",
                        test_dataset["idna_subject"],
                    )

                    PyFunceble.cli.utils.stdout.print_single_line("A")
                    return None

            if "from_inactive" not in test_dataset and self.inactive_dataset.exists(
                test_dataset
            ):
                # "I" means that it was ignored because of the inactive (db)
                # logic.

                PyFunceble.facility.Logger.info(
                    "Ignoring %r because it was already tested previously "
                    "(inactive).",
                    test_dataset["idna_subject"],
                )

                PyFunceble.cli.utils.stdout.print_single_line("I")

                return test_dataset, "ignored_inactive"

        PyFunceble.facility.Logger.info(
            "Started test of %r.",
            test_dataset["idna_subject"],
        )

        self._init_testing_object(
            test_dataset["subject_type"], test_dataset["checker_type"]
        )

        result = (
            self.testing_object.set_subject(test_dataset["idna_subject"])
            .query_status()
            .get_status()
        )

        PyFunceble.facility.Logger.info(
            "Successfully handled %r.",
            test_dataset["idna_subject"],
        )

        PyFunceble.facility.Logger.debug("Got status:\n%r.", result)

        return test_dataset, result



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the base of all syntax checker classes.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from typing import Optional

from sqlalchemy.orm import Session

import PyFunceble.storage
from PyFunceble.checker.base import CheckerBase
from PyFunceble.checker.syntax.params import SyntaxCheckerParams
from PyFunceble.checker.syntax.status import SyntaxCheckerStatus


class SyntaxCheckerBase(CheckerBase):
    """
    Provides the base of all our syntax checker classes.

    :param str subject:
        Optional, The subject to work with.
    """

    db_session: Optional[Session] = None

    status: Optional[SyntaxCheckerStatus] = None

    params: Optional[SyntaxCheckerParams] = None

    def __init__(
        self,
        subject: Optional[str] = None,
        db_session: Optional[Session] = None,
    ) -> None:
        self.params = SyntaxCheckerParams()

        self.status = SyntaxCheckerStatus()
        self.status.params = self.params

        super().__init__(subject=subject, db_session=db_session)

    def subject_propagator(self) -> "CheckerBase":
        """
        Propagate the currently set subject.

        .. warning::
            You are not invited to run this method directly.
        """

        self.status = SyntaxCheckerStatus()

        self.status.subject = self.subject
        self.status.idna_subject = self.idna_subject

        return self

    @CheckerBase.ensure_subject_is_given
    @CheckerBase.update_status_date_after_query
    def query_status(self) -> "SyntaxCheckerBase":
        """
        Queries the status.
        """

        if self.is_valid():
            self.status.status = PyFunceble.storage.STATUS.valid
        else:
            self.status.status = PyFunceble.storage.STATUS.invalid

        self.status.status_source = "SYNTAX"

        return self

    @CheckerBase.ensure_subject_is_given
    def is_valid(self) -> bool:
        raise NotImplementedError()

    # pylint: disable=useless-super-delegation
    def get_status(
        self,
    ) -> Optional[SyntaxCheckerStatus]:  # pragma: no cover ## Safety.
        return super().get_status()



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the our inactive DB migrator.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import datetime
import os
from typing import Optional

import domain2idna

import PyFunceble.cli.storage
import PyFunceble.facility
import PyFunceble.storage
from PyFunceble.cli.migrators.json2csv.base import JSON2CSVMigratorBase
from PyFunceble.cli.utils.stdout import print_single_line
from PyFunceble.cli.utils.testing import get_destination_from_origin
from PyFunceble.dataset.inactive.csv import CSVInactiveDataset
from PyFunceble.helpers.file import FileHelper


class InactiveJSON2CSVMigrator(JSON2CSVMigratorBase):
    """
    The migrator of the inactive database dile.
    """

    dataset: Optional[CSVInactiveDataset] = None

    def __post_init__(self) -> None:
        self.source_file = os.path.join(
            PyFunceble.storage.CONFIG_DIRECTORY,
            PyFunceble.cli.storage.INACTIVE_DB_OLD_FILE,
        )

        self.dataset = CSVInactiveDataset()

        return super().__post_init__()

    def migrate(self) -> "InactiveJSON2CSVMigrator":
        """
        Starts the migration.
        """

        file_helper = FileHelper(self.source_file)

        if file_helper.exists():
            self.dataset.set_authorized(True)
            dataset = {
                "idna_subject": None,
                "status": None,
                "status_source": None,
                "checker_type": "AVAILABILITY",
                "destination": None,
                "source": None,
                "tested_at": None,
                "session_id": None,
            }

            delete_file = True

            with file_helper.open("r", encoding="utf-8") as file_stream:
                for line in file_stream:
                    if (
                        self.continuous_integration
                        and self.continuous_integration.is_time_exceeded()
                    ):
                        delete_file = False
                        break

                    line = (
                        line.strip()
                        .replace('"', "")
                        .replace(",", "")
                        .replace(
                            "{",
                            "",
                        )
                        .replace("}", "")
                    )

                    if ":" not in line:
                        continue

                    index, value = [x.strip() for x in line.rsplit(":", 1)]

                    if not value:
                        if index.isdigit():
                            dataset["tested_at"] = datetime.datetime.fromtimestamp(
                                float(index)
                            ).isoformat()
                        else:
                            dataset["source"] = os.path.abspath(index)
                            dataset["destination"] = get_destination_from_origin(
                                dataset["source"]
                            )

                        continue

                    dataset["idna_subject"] = domain2idna.domain2idna(index)
                    dataset["status"] = value

                    if not dataset["tested_at"]:
                        dataset["tested_at"] = datetime.datetime.utcnow().isoformat()

                    PyFunceble.facility.Logger.debug("Decoded dataset:\n%r.", dataset)

                    self.dataset.update(dataset)

                    if self.print_action_to_stdout:
                        print_single_line()

                    PyFunceble.facility.Logger.info(
                        "Added %r into %r", dataset["idna_subject"], self.dataset
                    )

            if delete_file:
                file_helper.delete()
                self.done = True
        return self



# pylint:disable=invalid-name, cyclic-import
"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides a central storage place.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import os
import secrets
from typing import Optional

from box import Box
from dotenv import load_dotenv

from PyFunceble.storage_facility import get_config_directory

PROJECT_NAME: str = "PyFunceble"
PROJECT_VERSION: str = "4.1.3a2.dev (Blue Duckling: Ixora)"

DISTRIBUTED_CONFIGURATION_FILENAME: str = ".PyFunceble_production.yaml"
DISTRIBUTED_DIR_STRUCTURE_FILENAME: str = "dir_structure_production.json"

IANA_DUMP_FILENAME: str = "iana-domains-db.json"
PUBLIC_SUFFIX_DUMP_FILENAME: str = "public-suffix.json"
CONFIGURATION_FILENAME: str = ".PyFunceble.yaml"
CONFIGURATION_OVERWRITE_FILENAME: str = ".PyFunceble.overwrite.yaml"
ENV_FILENAME: str = ".pyfunceble-env"
DOWN_FILENAME: str = ".pyfunceble_intern_downtime.json"
USER_AGENT_FILENAME: str = "user_agents.json"
IPV4_REPUTATION_FILENAME: str = "ipv4_reputation.data"

# pylint: disable=line-too-long
IANA_DUMP_LINK: str = (
    "https://raw.githubusercontent.com/PyFunceble/iana/master/iana-domains-db.json"
)
PUBLIC_SUFFIX_DUMP_LINK: str = "https://raw.githubusercontent.com/PyFunceble/public-suffix/master/public-suffix.json"
USER_AGENT_DUMP_LINK: str = (
    "https://raw.githubusercontent.com/PyFunceble/user_agents/master/user_agents.json"
)
IPV4_REPUTATION_DUMP_LINK: str = "https://reputation.alienvault.com/reputation.data"

SHORT_REPO_LINK: str = "https://git.io/vpZoI"
REPO_LINK: str = "https://github.com/funilrys/PyFunceble"

NOT_RESOLVED_STD_HOSTNAME: str = f"pyfunceble-{secrets.token_hex(12)}.com"

IANA: Optional[dict] = {}
PUBLIC_SUFFIX: Optional[dict] = {}
USER_AGENTS: Optional[dict] = {}

load_dotenv(".env")
load_dotenv(ENV_FILENAME)

CONFIG_DIRECTORY: str = get_config_directory(
    project_name=PROJECT_NAME, project_version=PROJECT_VERSION
)
load_dotenv(os.path.join(CONFIG_DIRECTORY, ".env"))
load_dotenv(os.path.join(CONFIG_DIRECTORY, ENV_FILENAME))

# This is for the case that the environment variable are telling us something
# else.
CONFIG_DIRECTORY: str = get_config_directory(
    project_name=PROJECT_NAME, project_version=PROJECT_VERSION
)


CONFIGURATION: Optional[Box] = Box({})
FLATTEN_CONFIGURATION: Optional[dict] = {}
STATUS: Optional[Box] = Box(
    {
        "up": "ACTIVE",
        "down": "INACTIVE",
        "valid": "VALID",
        "invalid": "INVALID",
        "malicious": "MALICIOUS",
        "sane": "SANE",
    },
    frozen_box=True,
)
HTTP_CODES: Optional[Box] = Box({})
COLLECTION: Optional[Box] = Box({})
LINKS: Optional[Box] = Box({})
PROXY: Optional[Box] = Box({})


STD_HTTP_CODES: Optional[Box] = Box(
    {
        "list": {
            "up": [
                100,
                101,
                102,
                200,
                201,
                202,
                203,
                204,
                205,
                206,
                207,
                208,
                226,
                429,
            ],
            "potentially_up": [
                300,
                301,
                302,
                303,
                304,
                305,
                307,
                308,
                403,
                405,
                406,
                407,
                408,
                411,
                413,
                417,
                418,
                421,
                422,
                423,
                424,
                426,
                428,
                431,
                500,
                501,
                502,
                503,
                504,
                505,
                506,
                507,
                508,
                510,
                511,
            ],
            "potentially_down": [400, 402, 404, 409, 410, 412, 414, 415, 416, 451],
        },
    },
    frozen_box=True,
)
